﻿using Kivii.Lims.ExcelToPdf.Extensions;
using Kivii.Lims.ExcelToPdf.Requests;
using Kivii.Lims.RabbitMq.Entities;
using Kivii.Office;
using Kivii.RabbitMq;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Kivii.Lims.ExcelToPdf
{
    public partial class FrmServer : Form
    {
        #region 私有属性
        /// <summary>
        /// PDF上传客户端
        /// </summary>
        private JsonServiceClient _serviceClient = null;
        /// <summary>
        /// RabbitMq Server 
        /// </summary>
        private RabbitMqServer _mqServer = null;

        /// <summary>
        /// 默认配置文件路径，可以从服务器上存储目录配置AppConfig.json
        /// </summary>
        private string _defaultConfigPath = Path.Combine(KiviiContext.BaseDirectory, "AppConfig.json");
        /// <summary>
        /// 本地配置文件路径
        /// </summary>
        private string _localConfigPath = Path.Combine(KiviiContext.AppDataDirectory, "LocalConfig.json");
        /// <summary>
        /// 当前程序的配置信息
        /// </summary>
        private AppConfig _config = null;

        /// <summary>
        /// PDF签名模板
        /// </summary>
        private List<Stamper> _stampers = null;

        /// <summary>
        /// 下载报告客户端
        /// </summary>
        //private WebClient _downloadClient = null;

        /// <summary>
        /// Excel处理器
        /// </summary>
        //private OfficeApplication _application = null;
        #endregion

        #region 启停状态
        private bool _isProcessing = false;
        /// <summary>
        /// 服务正在处理中
        /// </summary>
        private bool isProcessing
        {
            get
            {
                return _isProcessing;
            }
            set
            {
                _isProcessing = value;
                this.Invoke(new Action(() =>
                {
                    btnConnect.Enabled = !_isProcessing;
                    if (!isStarted) lblThreads.Text = "";
                }));
            }
        }
        private bool _isStarted = false;
        /// <summary>
        /// 启动成功标记
        /// </summary>
        private bool isStarted
        {
            get
            {
                return _isStarted;
            }
            set
            {
                _isStarted = value;
                this.Invoke(new Action(() =>
                {
                    btnConnect.Text = value ? "停止服务" : "启动服务";
                    btnConnect.BackColor = value ? Color.Red : Color.LightSeaGreen;
                    lblThreads.Text = _isStarted ? _config.MqThreads.ToString() : "";
                }));
            }
        }
        #endregion

        #region 私有方法
        /// <summary>
        /// 输出消息
        /// </summary>
        /// <param name="message">消息内容</param>
        /// <param name="clearMessage">是否清除消息</param>
        private void outputMessage(string message, bool clearMessage = false)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new Action(() => outputMessage(message, clearMessage)));
                return;
            }
            if (clearMessage) rtbMessage.Clear();
            if (rtbMessage.Lines.Count() > 500) rtbMessage.Clear();
            var time = DateTime.Now.ToString("HH:mm:ss");
            rtbMessage.AppendText($"[{time}]{message}{Environment.NewLine}");
            rtbMessage.ScrollToCaret();
        }
        /// <summary>
        /// 加载配置信息，后期可放入配置文件中
        /// </summary>
        private AppConfig loadDefaultConfig()
        {
            if (this.InvokeRequired)
            {
                return this.Invoke(new Func<AppConfig>(() => readConfigFromUi())) as AppConfig;
            }
            AppConfig rtns = null;
            if (File.Exists(_localConfigPath))
            {
                var configString = File.ReadAllText(_localConfigPath);
                try
                {
                    rtns = configString.FromJson<AppConfig>();
                    return rtns;
                }
                catch
                { }
            }
            if (File.Exists(_defaultConfigPath))
            {
                var configString = File.ReadAllText(_defaultConfigPath);
                try
                {
                    rtns = configString.FromJson<AppConfig>();
                    return rtns;
                }
                catch
                { }
            }
            rtns = new AppConfig();
            rtns.MqAddress = "srv.kivii.org:5672";// "10.0.6.103";
            rtns.MqUserName = "guest";// "jst";
            rtns.MqPassword = "guest";// "www.kivii.com";
            rtns.MqTimeout = 10;
            rtns.MqRetry = true;
            rtns.MqThreads = 2;

            rtns.ServiceUrl = "";//"http://k5.jst-gov.com";
            rtns.QrCodeUrl = "";//"https://www.jst-gov.com";
            rtns.RouteStorage = "/Storages/Report/Excel01";
            rtns.ServiceUserName = "";//"administrator";
            rtns.ServicePassword = "";//"www.kivii.com";
            rtns.ServiceTimeout = 20;
            return rtns;
        }
        /// <summary>
        /// 读取界面上的值，因为在线程中，所以要通过方法来获取 
        /// </summary>
        /// <returns></returns>
        private AppConfig readConfigFromUi()
        {
            if (this.InvokeRequired)
            {
                return this.Invoke(new Func<AppConfig>(() => readConfigFromUi())) as AppConfig;
            }
            AppConfig rtns = new AppConfig();
            rtns.MqAddress = tbxMqAddress.Text;
            rtns.MqUserName = tbxMqUserName.Text;
            rtns.MqPassword = tbxMqPassword.Text;
            rtns.MqTimeout = (int)nudMqTimeout.Value;
            rtns.MqRetry = cbxMqRetry.Checked;
            rtns.MqThreads = (int)nudMqThreads.Value;

            rtns.ServiceUrl = tbxServiceUrl.Text;
            rtns.QrCodeUrl = tbxQRCodeUrl.Text;
            rtns.RouteStorage = tbxRouteStorage.Text;
            rtns.ServiceUserName = tbxServiceUserName.Text;
            rtns.ServicePassword = tbxServicePassword.Text;
            rtns.ServiceTimeout = (int)nudServiceTimeout.Value;
            return rtns;
        }
        /// <summary>
        /// 向界面写入配置信息
        /// </summary>
        /// <param name="config"></param>
        private void writeConfigToUi(AppConfig config)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new Action(() => writeConfigToUi(config)));
                return;
            }
            tbxMqAddress.Text = config.MqAddress;
            tbxMqUserName.Text = config.MqUserName;
            tbxMqPassword.Text = config.MqPassword;
            nudMqTimeout.Value = config.MqTimeout;
            cbxMqRetry.Checked = config.MqRetry;
            nudMqThreads.Value = config.MqThreads;

            tbxServiceUrl.Text = config.ServiceUrl;
            tbxServiceUserName.Text = config.ServiceUserName;
            tbxServicePassword.Text = config.ServicePassword;
            nudServiceTimeout.Value = config.ServiceTimeout;
            tbxRouteStorage.Text = config.RouteStorage;
            tbxQRCodeUrl.Text = config.QrCodeUrl;
        }
        #endregion

        public FrmServer()
        {
            InitializeComponent();
        }

        private void FrmServer_Load(object sender, EventArgs e)
        {
            _config = loadDefaultConfig();
            writeConfigToUi(_config);
            new Thread(new ThreadStart(tryStart)).Start();
        }

        #region 启动和停止程序步骤代码
        private void btnConnect_Click(object sender, EventArgs e)
        {
            if (isProcessing) return;
            if (isStarted)
            {
                new Thread(new ThreadStart(tryStop)).Start();
            }
            else
            {
                new Thread(new ThreadStart(tryStart)).Start();
            }
        }
        private void tryStart()
        {
            if (isProcessing) return;
            isProcessing = true;
            _config = readConfigFromUi();
            try
            {
                #region 前期预操作
                outputMessage("启动程序", true);
                if (_serviceClient != null)
                {
                    _serviceClient.Dispose();
                    _serviceClient = null;
                    outputMessage("Service Client已连接，重置连接");
                }
                if (_mqServer != null)
                {
                    outputMessage("MQ Server已连接，重置连接");
                    outputMessage("停止工作线程(20秒)...");
                    _mqServer.WaitForWorkersToStop(TimeSpan.FromSeconds(20));
                    _mqServer.Stop();
                    _mqServer.Dispose();
                    _mqServer = null;
                }
                #endregion

                #region 登陆服务器
                _serviceClient = new JsonServiceClient(_config.ServiceUrl);
                //设置请求超时
                _serviceClient.Timeout = TimeSpan.FromSeconds(_config.ServiceTimeout);
                try
                {
                    outputMessage($"开始登陆服务器:{_config.ServiceUrl}");
                    //定义一个身份认证数据包,因为登陆会有异常，所以要try,catch
                    var loginRequest = new Authenticate { UserName = _config.ServiceUserName, Password = _config.ServicePassword, provider = "Kivii", UseTokenCookie = true };
                    //因为在线程中，所以不需要再用异步了
                    var loginResponse = _serviceClient.Post(loginRequest);
                    outputMessage($"登陆成功:{loginResponse.DisplayName}");
                }
                catch (Exception ex)
                {
                    outputMessage($"登陆失败:{ex.Message}");
                    _serviceClient.Dispose();
                    _serviceClient = null;
                    return;
                }
                #endregion

                #region 获取PDF签名模板
                try
                {
                    outputMessage($"开始获取PDF签名模板...");
                    var stamperRequest = new StamperQuery();
                    var stamperResponse = _serviceClient.Post(stamperRequest);
                    _stampers = stamperResponse.Results;
                    _stampers.IsNullOrEmpty().ThrowIfTrue("未能查询到PDF签名模板");
                    outputMessage($"获取PDF签名模板成功:{_stampers.Select(o => o.Name).Join(",")}");
                }
                catch (Exception ex)
                {
                    outputMessage($"获取PDF签名模板失败:{ex.Message}");
                    _serviceClient.Dispose();
                    _serviceClient = null;
                    return;
                }
                #endregion

                #region 连接MQ并注册服务
                try
                {
                    outputMessage($"开始连接MQ服务器:{_config.MqAddress}");
                    var factory = new RabbitMqMessageFactory(_config.MqAddress, _config.MqUserName, _config.MqPassword);
                    factory.ConnectionFactory.RequestedConnectionTimeout = _config.MqTimeout * 1000;
                    //测试连接，如果连接不上，异常
                    var conn = factory.ConnectionFactory.CreateConnection();
                    conn.Dispose();
                    outputMessage($"连接MQ服务器成功");
                    outputMessage($"注册消息服务...");
                    _mqServer = new RabbitMqServer(factory);
                    _mqServer.RetryCount = _config.MqRetry ? 1 : 0;//如果处理过程中发生异常，重试次数，重新发送,只能是1或0
                    _mqServer.RegisterHandler<ReportMqRequest>(processReportMqRequest, _config.MqThreads);
                    _mqServer.Start();
                    outputMessage($"注册消息服务成功，运行线程数:{_config.MqThreads}");
                    isStarted = true;
                }
                catch (Exception ex)
                {
                    try
                    {
                        _mqServer.MessageFactory.Dispose();
                    }
                    catch { }
                    outputMessage($"连接MQ失败:{ex.Message}");
                    _mqServer = null;
                }
                #endregion
            }
            catch (Exception ex)
            {
                outputMessage($"发生未捕获的错误:{ex.Message}");
                isStarted = false;
            }
            finally
            {
                isProcessing = false;
            }
        }
        private void tryStop()
        {
            if (isProcessing) return;
            isProcessing = true;

            try
            {
                outputMessage($"正在停止Service Client连接...");
                _serviceClient?.Dispose();
                _serviceClient = null;
                outputMessage($"正在停止MQ Server...");
                _mqServer?.Stop();
                outputMessage($"正在停止MQ 连接...");
                _mqServer?.MessageFactory?.Dispose();
                _mqServer?.Dispose();
                _mqServer = null;
            }
            catch (Exception ex)
            {
                outputMessage($"发生未捕获的错误:{ex.Message}");
            }
            finally
            {
                outputMessage($"服务已停止");
                isProcessing = false;
                isStarted = false;
            }
        }
        #endregion

        #region 按钮交互
        private void btnEditConfig_Click(object sender, EventArgs e)
        {
            if (!gbMqConfig.Enabled)
            {
                gbMqConfig.Enabled = true;
                gbServiceConfig.Enabled = true;
                btnEditConfig.Text = "保存配置";
                return;
            }
            //以下是保存动作
            gbMqConfig.Enabled = false;
            gbServiceConfig.Enabled = false;
            btnEditConfig.Text = "编辑配置";
            var config = readConfigFromUi();
            var configJson = config.ToJson();
            try
            {
                File.WriteAllText(_localConfigPath, configJson, Encoding.UTF8);
            }
            catch (UnauthorizedAccessException ex)
            {
                outputMessage($"权限不足,无法更改配置信息,请使用管理员身份运行后重试!:\n{ex}", true);
            }
        }

        private void cmsNotify_Exit_Click(object sender, EventArgs e)
        {
            try
            {
                notify.Dispose();
                tryStop();
            }
            catch
            { }
            finally
            {
                System.Environment.Exit(0);
            }
        }

        private void FrmServer_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = true;
            this.Visible= false;
            notify.Visible = true;
            //this.Hide();
            notify.BalloonTipText = "后台运行";
            notify.Text = $"{this.Text}-当前地址：{tbxServiceUrl.Text}";
            notify.ShowBalloonTip(1000);
        }

        private void notify_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            this.Visible = true;
            this.Activate();
            notify.Visible = false;
            this.Show();
        }
        #endregion

        #region 业务方法
        private object processReportMqRequest(Kivii.Messaging.IMessage<ReportMqRequest> message)
        {
            var request = message.GetBody();
            outputMessage($"处理请求:{request.ReportId},RetryAttempts={message.RetryAttempts},消息Id={message.Id}");
            if (ExcuteExcel(request))
            {
                CreatePrintPool(request);
                outputMessage($"处理结束:{request.ReportId}");
                Thread.Sleep(1000);
                return null;
            }
            else
            {
                return request;
                //throw new Exception(request.Exception);
            }
        }

        /// <summary>
        /// 处理报告Excel
        /// </summary>
        /// <param name="reportMqRequest"></param>
        /// <returns></returns>
        private bool ExcuteExcel(ReportMqRequest reportMqRequest)
        {
            OfficeApplication application = null;
            try
            {
                var downloadClient = new WebClient();

                var tempDownloadFilePath = KiviiContext.GetTemporaryFilePath($"{reportMqRequest.ReportId}.xlsx");
                var tempPDFFilePath = KiviiContext.GetTemporaryFilePath($"{reportMqRequest.ReportId}.pdf");
                var downloadUrl = string.Empty;
                if (reportMqRequest.OpenFilePath.StartsWithIgnoreCase("http")) downloadUrl = reportMqRequest.OpenFilePath;
                else if (reportMqRequest.OpenFilePath.StartsWithIgnoreCase("/")) downloadUrl = $"{_config.ServiceUrl}{reportMqRequest.OpenFilePath}";
                else downloadUrl = $"{_config.ServiceUrl}{_config.RouteStorage}/{reportMqRequest.Kvid}/{reportMqRequest.ReportId}.xlsx";
                downloadClient.DownloadFile(downloadUrl, tempDownloadFilePath);
                if (!File.Exists(tempDownloadFilePath))
                {
                    outputMessage($"失败:未找到要下载的excel文件!");
                    outputMessage($"失败:处理对象-> Kvid:{reportMqRequest.Kvid} ReportId:{reportMqRequest.ReportId} Stamp:{reportMqRequest.Stamp}");
                    return false;
                }
                outputMessage($"报告({reportMqRequest.ReportId})excel文件下载成功!");
                application = OfficeApplication.CreateNewApplication(tempDownloadFilePath);
                application.Visible = false;
                application.DisplayAlerts = false;
                application.Open();
                outputMessage($"报告({reportMqRequest.ReportId})excel已打开!");
                application.SaveAs(tempPDFFilePath);
                outputMessage($"报告({reportMqRequest.ReportId})已经合成pdf!");
                application.Close();
                //_application.Dispose();
                outputMessage($"报告({reportMqRequest.ReportId})excel已经关闭!");
                if (!File.Exists(tempPDFFilePath))
                {
                    outputMessage($"失败:未找到转后的PDF文件!");
                    outputMessage($"失败:处理对象-> Kvid:{reportMqRequest.Kvid} ReportId:{reportMqRequest.ReportId} Stamp:{reportMqRequest.Stamp}");
                    return false;
                }

                var serviceClient = new JsonServiceClient(_config.ServiceUrl);
                outputMessage($"报告({reportMqRequest.ReportId})准备调用PDFStamper!");

                var code = reportMqRequest.Kvid.ToString().Substring(0, 4);
                var internalCode = $"{code}/{reportMqRequest.ReportId}";
                var redirectUrl = $"{_config.QrCodeUrl}/Url/{internalCode}";

                var stampInternalCode = "新中文印章";// reportMqRequest.Stamp;

                if(_stampers.Exists(o => o.InternalCode == reportMqRequest.Stamp)) stampInternalCode =  reportMqRequest.Stamp;
                //else if (_stampers.Exists(o => o.Name.Contains(reportMqRequest.Stamp)))
                else if (_stampers.Exists(o => reportMqRequest.Stamp.Contains(o.Name)))
                        {
                    //var _currentStamp = _stampers.FirstOrDefault(o=>o.Name.Contains(reportMqRequest.Stamp));
                    var _currentStamp = _stampers.FirstOrDefault(o => reportMqRequest.Stamp.Contains(o.Name));
                    if (_currentStamp != null)
                    {
                        stampInternalCode = _currentStamp.InternalCode;
                    }
                }

                var request = new StamperRequest();
                request.InternalCode = stampInternalCode;
                request.OpenFilePath = $"{reportMqRequest.ReportId}.pdf";
                request.SaveFilePath = reportMqRequest.SaveFilePath.IsNullOrEmpty() ? $"/Storages/Reports/Pdf/{reportMqRequest.Kvid}/{reportMqRequest.ReportId}.pdf" : reportMqRequest.SaveFilePath;
                request.StampFiles = new Dictionary<string, string>();
                request.StampFiles.Add("BarCode", $"/Barcode?Format=QR_CODE&Code={redirectUrl}");
                var files = new List<UploadFile>();
                files.Add(new UploadFile($"{reportMqRequest.ReportId}.pdf", new FileStream(tempPDFFilePath, FileMode.Open)));

                var response = serviceClient.PostFilesWithRequest<StamperResponse>(request, files);
                outputMessage($"报告({reportMqRequest.ReportId})已调用PDFStamper:{stampInternalCode}!");
                files.ForEach(o =>
                {
                    o.Stream.Close();
                    o.Stream.Dispose();
                });
                if (response == null)
                {
                    outputMessage($"失败:请求PDF处理器(Stamper)未响应!");
                    outputMessage($"失败:处理对象-> Kvid:{reportMqRequest.Kvid} ReportId:{reportMqRequest.ReportId} Stamp:{reportMqRequest.Stamp}");
                    return false;
                }
                if (!response.IsSucceed)
                {

                    outputMessage($"失败:请求PDF处理器(Stamper)未响应!");
                    outputMessage($"失败:响应内容: {response.Message}!");
                    outputMessage($"失败:处理对象-> Kvid:{reportMqRequest.Kvid} ReportId:{reportMqRequest.ReportId} Stamp:{request.InternalCode}!");
                }
                return response.IsSucceed;
            }
            catch (Exception ex)
            {
                reportMqRequest.Exception = ex.Message;
                if (application != null)
                {
                    application.Close();
                    application.Dispose();
                    application = null;
                }
                //_application.Dispose();
                this.Invoke(new Action(() =>
                {
                    outputMessage($"ExcuteExcel:Error,{ex.Message}!");
                    outputMessage($"ExcuteExcel: 处理对象-> Kvid:{reportMqRequest.Kvid} ReportId:{reportMqRequest.ReportId} Stamp:{reportMqRequest.Stamp}!");
                }));
                return false;
            }
            finally
            {
                if (application != null)
                {
                    application.Close();
                    application.Dispose();
                    application = null;
                }
            }
        }

        private bool CreatePrintPool(ReportMqRequest reportMqRequest)
        {
            try
            {
                var code = reportMqRequest.Kvid.ToString().Substring(0, 4);
                var internalCode = $"{code}/{reportMqRequest.ReportId}";
                var printPool = new PrintPoolRequest();
                printPool.Item = new PrintPool();
                printPool.Item.OwnerKvid = reportMqRequest.Kvid;
                printPool.Item.OwnerName = reportMqRequest.DemanderName;
                printPool.Item.Title = reportMqRequest.ReportId;
                printPool.Item.Category = "Kivii.Lims.Entities.Report";//区别于新老Lims
                printPool.Item.Count = (int)reportMqRequest.Quantity;
                if (reportMqRequest.Quantity <= 0)
                {
                    printPool.Item.IsPrint = true;
                    printPool.Item.Result = "标记为已打印";
                }
                printPool.Item.OperateTime = reportMqRequest.IssueDate == null ? DateTime.Now : reportMqRequest.IssueDate.Value;//签发日期
                printPool.Item.OperatorName = reportMqRequest.PreparerName.IsNullOrEmpty() ? reportMqRequest.IssuerName : reportMqRequest.PreparerName;
                printPool.Item.OperatorKvid = reportMqRequest.PreparerKvid == Guid.Empty ? reportMqRequest.IssuerKvid : reportMqRequest.PreparerKvid;
                printPool.Item.FilePath = $"{_config.QrCodeUrl}/Lims/Report/{internalCode}";
                printPool.Item.Summary = reportMqRequest.TestJudgement;//判定
                printPool.Item.Remark = reportMqRequest.Conclusion;//新lims 中检验结论
                printPool.Item.Metadata = new Dictionary<string, string>();
                printPool.Item.Metadata = reportMqRequest.Metadata;
                foreach (var item in reportMqRequest.GetType().GetProperties())
                {
                    try
                    {
                        var key = item.Name;
                        if (key == "Metadata") continue;
                        var type = reportMqRequest.GetType().GetProperty(key).GetType();
                        var value = reportMqRequest.GetType().GetProperty(key).GetValue(reportMqRequest, null);
                        if (value == null) continue;
                        printPool.Item.Metadata[key] = value.ToString();
                        if (type == typeof(DateTime)) printPool.Item.Metadata[key] = DateTime.Parse(value.ToString()).ToString("yyyy-MM-dd");
                    }
                    catch { continue; }
                }
                var serviceClient = new JsonServiceClient(tbxServiceUrl.Text);
                serviceClient.Post(printPool);
                outputMessage($"CreatePrintPool:推动打印成功:{printPool.Item.Title}!");
                //rtbMessage.AppendText($"CreatePrintPool: 推动打印成功:{printPool.Item.Title}\n");
                return true;
            }
            catch (Exception ex)
            {
                outputMessage($"CreatePrintPool:推动打印失败:{ex.Message}!");
                //rtbMessage.AppendText($"CreatePrintPool: 推动打印失败:{ex.Message}\n");
                return false;
            }
        }
        #endregion
    }
}
